/******************************************************************************
 *
 * Copyright (c) 1998,99 by Mindbright Technology AB, Stockholm, Sweden.
 *                 www.mindbright.se, info@mindbright.se
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 *****************************************************************************
 * $Author: mats $
 * $Date: 1999/01/31 19:00:36 $
 * $Name: rel0-98-1 $
 *****************************************************************************/
package mindbright.application;

import java.io.*;

import java.awt.*;
import java.awt.event.*;
import java.util.NoSuchElementException;

import mindbright.ssh.*;
import mindbright.security.*;
import mindbright.terminal.*;

public class MindTerm {

  Frame       frame;
  MenuBar     menubar;
  TerminalWin term;

  SSHInteractiveClient client;
  SSHStdIO             console;
  
  public MindTerm() {
  }

  public static void main(String[] argv) {
    MindTerm controller = new MindTerm();
    controller.startup(argv);
  }

  public void startup(String[] argv) {
    try {
      String  server         = null;
      String  commandLine    = null;
      String  sshHomeDir     = null;
      boolean haveGUI        = true;
      boolean usePropFiles   = true;
      boolean useDefPropOnly = false;
      int     i, numOfOpts;

      client  = new SSHInteractiveClient(System.getProperty("user.home", "~") + File.separator +
					 ".ssh" + File.separator,
					 System.getProperty("user.name", ""));
      console = (SSHStdIO)client.getConsole();

      // First we check the MindTerm options (i.e. not the ssh/terminal-properties)
      //
      try {
	for(i = 0; i < argv.length; i++) {
	  String arg = argv[i];
	  if(!arg.startsWith("--"))
	    break;
	  switch(arg.charAt(2)) {
	  case 'h':
	    sshHomeDir = argv[++i];
	    break;
	  case 'd':
	    haveGUI = false;
	    break;
	  case 'p':
	    String typ = argv[++i];
	    if(typ.equals("no"))
	      usePropFiles = false;
	    else if(typ.equals("def"))
	      useDefPropOnly = true;
	    else
	      throw new Exception();
	    break;
	  case 'v':
	    System.out.println("verbose mode selected...");
	    SSH.DEBUG = true;
	    break;
	  case 'V':
	    System.out.println("SSH version " + SSH.VER_SSH + ", protocol version " +
			       SSH.VER_MAJOR + "." + SSH.VER_MINOR);
	    System.out.println(SSH.VER_MINDTERM);
	    System.exit(0);
	    break;
	  default:
	    throw new Exception();
	  }
	}
      } catch (Exception e) {
	System.out.println("usage: MindTerm [options] [properties] [command]");
	System.out.println("Options:");
	System.out.println("  --d            No terminal-window, only dumb command-line and port-forwarding.");
	System.out.println("  --h dir        Name of the MindTerm home-dir (default: ~/.ssh/).");
	System.out.println("  --p <no | def> Use no property-files or only default property-file.");
	System.out.println("  --v            Verbose; display verbose debugging messages.");
	System.out.println("  --V            Display version number only.");
	throw new Exception();
      }
      numOfOpts = i;

      if(sshHomeDir != null) {
	client.setSSHHomeDir(sshHomeDir);
      }

      // Must set these here (since we start setting properties below)
      //
      client.setUsePropertyFiles(usePropFiles);
      client.setUseOnlyDefaultProperties(useDefPropOnly);

      // Check to see if a server was given, in that case we first try
      // to load its properties
      //
      for(i = numOfOpts; i < argv.length; i++) {
	String arg = argv[i];
	if(argv[i].equals("-server"))
	  server = argv[i + 1];
      }

      // First we initialize the GUI if we have one (to be able to set
      // properties to terminal
      //
      if(haveGUI) {
	initGUI();
	console.setTerminal(term);
	console.setOwnerContainer(frame);
	console.setOwnerName(SSH.VER_MINDTERM);
	console.updateTitle();
      }

      client.printCopyright();

      if(usePropFiles) {
	// Load default or given server's properties
	//
	if(server == null || (!useDefPropOnly && !client.loadProperties(server))) {
	  if(!client.loadDefaultProperties()) {
	    console.println("First time using " + SSH.VER_MINDTERM + ", Welcome!");
	    console.println("Please read the file 'README' for instructions on usage.");
	    console.println("(visit <http://www.mindbright.se/mindterm/ for more information)");
	    String ans = console.promptLine("Do you want to save default properties to file (yes/no)? ", "", false);
	    if(ans.equalsIgnoreCase("yes") || ans.equalsIgnoreCase("y")) {
	      if(!client.saveDefaultProperties()) {
		console.println("Could not write default properties in '" +
				client.getSSHHomeDir() + "'.");
		usePropFiles = false;
	      } else {
		console.println("Your default properties were saved to '" +
				client.getSSHHomeDir() + "mindterm_def" + SSH.PROPS_FILE_EXT + "'.");
	      }
	    } else {
	      usePropFiles = false;
	    }
	  }
	}
      }

      // Disable use of property-files if asked for or if they can't be used
      //
      if(!usePropFiles) {
	console.println("Use of property-files disabled in this session!");
	client.setUsePropertyFiles(usePropFiles);
      } else if(useDefPropOnly) {
	console.println("Using only default property-file, all changes will be saved as default.");
      }

      // Check to see if there are any properties given on the command line (-<name> <value>)
      // if so, we merge them into the defaults (i.e. overriding defaults)
      //
      String name;
      String value;
      for(i = numOfOpts; i < argv.length; i += 2) {
	name = argv[i];
	if((name.charAt(0) != '-') || ((i + 1) == argv.length))
	  break;
	name  = name.substring(1);
	value = argv[i + 1];
	try {
	  try {
	    client.setProperty(name, value);
	  } catch (NoSuchElementException e1) {
	    if(term != null) {
	      try {
		term.setProperty(name, value);
	      } catch (NoSuchElementException e2) {
		console.println("No such property: " + name);
	      }
	    }
	  }
	} catch (IllegalArgumentException e3) {
	  console.println(e3.getMessage());
	}
      }

      if(i < argv.length) {
	commandLine = "";
	for(; i < argv.length; i++) {
	  commandLine += argv[i] + " ";
	}
	haveGUI = false;
      }

      if(commandLine != null) {
	try {
	  client.doSingleCommand(commandLine, false, 0);
	} catch (IOException e) {
	  System.out.println("Single command error: "  + e.getMessage());
	}
      } else {
	// Adhoc... Let the AWT-stuff settle before we try to print to the term...
	// (TODO: get rid of this, should wait for something more specific I guess :-)
	//
	Thread.sleep(1000);
	Thread clientThread = new Thread(client);
	clientThread.start();
	clientThread.join();
	windowClosing(null);
      }

    } catch (Exception e) {

      // !!! DEBUG !!!
      if(SSH.DEBUG) {
	System.out.println("Error: " + e);
	e.printStackTrace();
      }
      windowClosing(null);
    }
  }

  public void initGUI() {
    frame = new Frame();
    term  = new TerminalWin(frame, 80, 24, new TerminalXTerm());

    frame.addWindowListener(new WindowAdapter() {
      public void windowClosing(WindowEvent e)  { MindTerm.this.windowClosing(e); }

      // !!! I don't get it, why do this occur? Who gets focus when we
      // are deiconified??  (this do only apply to winDOS)
      //
      // NOTE: Without this AND the FocusListener.focusLost() in
      // TerminalWin.java we don't regain keyboard-focus after having
      // been iconified !!!
      //
      // NOTE2: This seems to be fixed by the fact that the Canvas is
      // not put directly into the Frame, but instead into a Panel
      // which is put there... (bug, feature, or am I a moron?)
      //
      public void windowDeiconified(WindowEvent e)  { term.requestFocus(); }
    });

    // !!! Weird, on Linux if the menubar is added here before the window is visible the
    // TerminalWin won't get its events... TODO: check this
    //
    // frame.setMenuBar(SSHMenuHandler.createMenubar());

    term.addMenuItem(null);
    SSHMenuHandler menus = new SSHMenuHandler(client, frame, term);
    term.addMenuItem(menus.getMindTermMenu());
    term.addMenuItem(menus.getSettingsMenu());
    term.addMenuItem(menus.getTunnelsMenu());

    frame.add(term.getPanelWithScrollbar());

    frame.pack();
    frame.show();
  }

  public void windowClosing(WindowEvent e) {
    client.saveCurrentProperties();
    System.out.println("Thank you for using MindTerm...");
    if(frame != null)
      frame.dispose();
    System.exit(0);
  }

}

