/******************************************************************************
 *
 * Copyright (c) 1998,99 by Mindbright Technology AB, Stockholm, Sweden.
 *                 www.mindbright.se, info@mindbright.se
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 *****************************************************************************
 * $Author: mats $
 * $Date: 1999/01/31 19:00:36 $
 * $Name: rel0-98-1 $
 *****************************************************************************/
package mindbright.ssh;

import java.net.*;
import java.io.*;
import java.math.BigInteger;
import java.util.Vector;

import mindbright.security.*;
import mindbright.terminal.*;

/**
 * This class contains the main functionality for setting up a
 * connection to a ssh-server. It can be used both to implement a
 * "full" ssh-client, or it can be used to fire off a singe command on
 * the server (both in a background thread and in the
 * current-/foreground-thread). A set of properties can be used to
 * control different aspects of the connection. These are fetched from
 * an object implementing the <code>SSHClientUser</code>-interface.
 * The authentication can be done in different ways, all which is
 * handled through an object implementing the
 * <code>SSHAuthenticator</code>-interface. The console-output of the
 * <code>SSHClient</code> is (optionally) handled through an object
 * implementing the <code>SSHConsole</code>-interface.
 * <p>
 *
 * A class realizing a full interactive ssh-client is
 * <code>SSHInteractiveClient</code>. The <code>SSHClient</code>-class
 * is also used transparently from the <code>SSHSocket</code>- and <code>SSHServerSocket</code>-
 * classes (through the <code>SSHSocketFactory</code>- and  <code>SSHSocketImpl</code>-classes).
 *
 * @author  Mats Andersson
 * @version 0.96, 26/11/98
 * @see     SSHAuthenticator
 * @see     SSHClientUser
 * @see     SSHConsole 
 * @see     SSHSocketFactory
 * @see     SSHSocketImpl
 */
public class SSHClient extends SSH {

  static public class ExitMonitor implements Runnable {
    SSHClient client;
    long      msTimeout;
    public ExitMonitor(SSHClient client, long msTimeout) {
      this.msTimeout  = msTimeout;
      this.client     = client;
    }
    public ExitMonitor(SSHClient client) {
      this(client, 0);
    }
    public void run() {
      client.waitForExit(msTimeout);
      // If we have allready exited gracefully don't report...
      //
      if(!client.gracefulExit)
	client.disconnect(false);
    }
  }

  // Local port forwarding
  //
  public static class LocalForward {
    protected int    localPort;
    protected String remoteHost;
    protected int    remotePort;
    protected String plugin;
    public LocalForward(int localPort, String remoteHost, int remotePort, String plugin) {
      this.localPort  = localPort;
      this.remoteHost = remoteHost;
      this.remotePort = remotePort;
      this.plugin     = plugin;
    }
  }

  // Remote port forwarding
  //
  public static class RemoteForward {
    protected int    remotePort;
    protected String localHost;
    protected int    localPort;
    protected String plugin;
    public RemoteForward(int remotePort, String localHost, int localPort, String plugin) {
      this.remotePort = remotePort;
      this.localHost  = localHost;
      this.localPort  = localPort;
      this.plugin     = plugin;
    }
  }

  protected InetAddress serverAddr;
  protected InetAddress localAddr;
  protected String      srvVersion;

  protected Vector localForwards;
  protected Vector remoteForwards;
  protected String commandLine;

  protected SSHChannelController controller;
  protected SSHConsole           console;
  protected SSHAuthenticator     authenticator;
  protected SSHClientUser        user;

  protected Socket               sshSocket;
  protected BufferedInputStream  sshIn;
  protected BufferedOutputStream sshOut;

  protected boolean gracefulExit;
  protected boolean isConnected;

  protected int refCount;

  public SSHClient(SSHAuthenticator authenticator, SSHClientUser user) {
    this.user           = user;
    this.authenticator  = authenticator;
    this.localForwards  = new Vector();
    this.remoteForwards = new Vector();
    this.srvVersion     = null;
    this.refCount       = 0;
  }

  public void setConsole(SSHConsole console) {
    this.console = console;
  }

  public SSHConsole getConsole() {
    return console;
  }

  public InetAddress getServerAddr() {
    return serverAddr;
  }

  public InetAddress getLocalAddr() {
    return localAddr;
  }

  public String getServerVersion() {
    return srvVersion;
  }

  public void addLocalPortForward(String fwdSpec, boolean commit) throws Exception {
    int    localPort;
    String remoteHost;
    int    remotePort;
    int    d1, d2, pd;
    String tmp, plugin;

    if(fwdSpec.charAt(0) == '/') {
      int i = fwdSpec.lastIndexOf('/');
      if(i == 0)
	throw new Exception("Invalid port forward spec.");
      plugin = fwdSpec.substring(1, i);
      fwdSpec = fwdSpec.substring(i + 1);
    } else
      plugin = "default";

    d1 = fwdSpec.indexOf(':');
    d2 = fwdSpec.lastIndexOf(':');
    if(d1 == d2)
      throw new Exception("Invalid port forward spec.");

    tmp        = fwdSpec.substring(0, d1);
    localPort  = Integer.parseInt(tmp);
    remoteHost = fwdSpec.substring(d1 + 1, d2);
    tmp        = fwdSpec.substring(d2 + 1);
    remotePort = Integer.parseInt(tmp);
    if(commit) {
      delLocalPortForward(localPort);
      addLocalPortForward(localPort, remoteHost, remotePort, plugin);
    }
  }

  public void addLocalPortForward(int localPort, String remoteHost, int remotePort, String plugin) {
    localForwards.addElement(new LocalForward(localPort, remoteHost, remotePort, plugin));
    if(isConnected)
      requestLocalPortForward(localPort, remoteHost, remotePort, plugin);
  }

  public void delLocalPortForward(int port) {
    if(port == -1) {
      if(isConnected)
	controller.killListenChannels();
      localForwards = new Vector();
    } else {
      for(int i = 0; i < localForwards.size(); i++) {
	LocalForward fwd = (LocalForward) localForwards.elementAt(i);
	if(fwd.localPort == port) {
	  localForwards.removeElementAt(i);
	  if(isConnected)
	    controller.killListenChannel(fwd.localPort);
	  break;
	}
      }
    }
  }

  public int getLocalPortForwardPort(int n) {
    return ((LocalForward)localForwards.elementAt(n)).localPort;
  }

  public void addRemotePortForward(String fwdSpec, boolean commit) throws Exception {
    int    remotePort;
    int    localPort;
    String localHost;
    int    d1, d2;
    String tmp, plugin;

    if(fwdSpec.charAt(0) == '/') {
      int i = fwdSpec.lastIndexOf('/');
      if(i == 0)
	throw new Exception("Invalid port forward spec.");
      plugin = fwdSpec.substring(1, i);
      fwdSpec = fwdSpec.substring(i + 1);
    } else
      plugin = "default";

    d1 = fwdSpec.indexOf(':');
    d2 = fwdSpec.lastIndexOf(':');
    if(d1 == d2)
      throw new Exception("Invalid port forward spec.");

    tmp        = fwdSpec.substring(0, d1);
    remotePort = Integer.parseInt(tmp);
    localHost  = fwdSpec.substring(d1 + 1, d2);
    tmp        = fwdSpec.substring(d2 + 1);
    localPort  = Integer.parseInt(tmp);
    if(commit) {
      delRemotePortForward(remotePort);
      addRemotePortForward(remotePort, localHost, localPort, plugin);
    }
  }

  public void addRemotePortForward(int remotePort, String localHost, int localPort, String plugin) {
    remoteForwards.addElement(new RemoteForward(remotePort, localHost, localPort, plugin));
  }

  public void delRemotePortForward(String plugin) {
    for(int i = 0; i < remoteForwards.size(); i++) {
      RemoteForward fwd = (RemoteForward) remoteForwards.elementAt(i);
      if(fwd.plugin.equals(plugin)) {
	remoteForwards.removeElementAt(i);
	if(isConnected)
	  controller.addHostMapPermanent(fwd.localHost, "127.0.0.1", 0);
      }
    }
  }

  public void delRemotePortForward(int port) {
    if(port == -1) {
      if(isConnected) {
	for(int i = 0; i < remoteForwards.size(); i++) {
	  RemoteForward fwd = (RemoteForward) remoteForwards.elementAt(i);
	  controller.addHostMapPermanent(fwd.localHost, "127.0.0.1", 0);
	}
      }
      remoteForwards = new Vector();
    } else {
      for(int i = 0; i < remoteForwards.size(); i++) {
	RemoteForward fwd = (RemoteForward) remoteForwards.elementAt(i);
	if(fwd.remotePort == port) {
	  remoteForwards.removeElementAt(i);
	  if(isConnected)
	    controller.addHostMapPermanent(fwd.localHost, "127.0.0.1", 0);
	  break;
	}
      }
    }
  }

  public int getRemotePortForwardPort(int n) {
    return ((RemoteForward)remoteForwards.elementAt(n)).remotePort;
  }

  public void startExitMonitor() {
    startExitMonitor(0);
  }

  public void startExitMonitor(long msTimeout) {
    (new Thread(new ExitMonitor(this, msTimeout))).start();
  }

  public synchronized int addRef() {
    return ++refCount;
  }

  public void forcedDisconnect() {
    controller.sendDisconnect("exit");
  }

  public synchronized int delRef() {
    if(--refCount <= 0) {
      forcedDisconnect();
      waitForExit(2000);
    }
    return refCount;
  }

  public void waitForExit(long msTimeout) {
    try {
      controller.waitForExit(msTimeout);
    } catch(InterruptedException e) {
      // !!!
      SSH.log("Error when shutting down SSHClient: " + e.getMessage());
      controller.killAll();
    }
  }

  public void doSingleCommand(String commandLine, boolean background, long msTimeout)
    throws IOException {
    this.commandLine = commandLine;
    bootSSH(false);
    if(background)
      startExitMonitor(msTimeout);
    else
      waitForExit(msTimeout);
  }

  public void bootSSH(boolean haveCnxWatch) throws IOException {
    try {
      // !!!
      initiatePlugins(haveCnxWatch);

      serverAddr = InetAddress.getByName(user.getSrvHost());
      localAddr  = InetAddress.getLocalHost();

      if(user.wantPrivileged()) {
	int p;
	for(p = 1023; p > 512; p--) {
	  try {
	    sshSocket  = new Socket(serverAddr, user.getSrvPort(), localAddr, p);
	  } catch (IOException e) {
	    if(e.getMessage().toLowerCase().indexOf("use") == -1)
	      throw e;
	    continue;
	  }
	  break;
	}
	if(p == 512)
	  throw new IOException("No available privileged ports");
      } else {
	sshSocket  = new Socket(serverAddr, user.getSrvPort());
      }
      sshIn  = new BufferedInputStream(sshSocket.getInputStream(), 8192);
      sshOut = new BufferedOutputStream(sshSocket.getOutputStream());

      negotiateVersion();

      // Get username before connected since interactive client uses
      // the connected() call-back to stop interacting (i.e. using
      // "saved" values)
      //
      String userName = authenticator.getUsername(user);

      // We now have a physical connection to a sshd, report this to the SSHClientUser
      //
      isConnected = true;
      user.connected(this);

      receiveServerData();

      cipherType = authenticator.getCipher(user);

      // Check that selected cipher is supported by server
      //
      if(!isCipherSupported(cipherType))
	throw new IOException("Sorry, server does not support the '" +
			      getCipherName(authenticator.getCipher(user)) + "' cipher.");

      generateSessionId();
      generateSessionKey();

      initClientCipher();

      sendSessionKey(cipherType);

      // !!!
      // At this stage the communication is encrypted
      // !!!

      authenticateUser(userName);

      controller = new SSHChannelController(this, sshIn, sshOut, sndCipher, rcvCipher,
					    console, haveCnxWatch);
      initiateSession();
      if(console != null)
	console.serverConnect(controller, sndCipher);
      controller.start();

    } catch (IOException e) {
      if(sshSocket != null)
	sshSocket.close();
      disconnect(false);
      throw e;
    }
  }

  protected void disconnect(boolean graceful) {
    if(!isConnected)
      return;
    isConnected  = false;
    gracefulExit = graceful;
    user.disconnected(this, graceful);
  }

  void negotiateVersion() throws IOException {
    byte[] buf = new byte[256];
    int    len;
    String verStr;

    len = sshIn.read(buf);

    // !!! For now...
    srvVersion = new String(buf, 0, len);
    if(srvVersion.indexOf("SSH-1.5") != 0 && srvVersion.indexOf("SSH-1.99") != 0)
      throw new IOException("Version mismatch, can't talk to server");

    // Strip trailing newline
    srvVersion = srvVersion.substring(0, srvVersion.length() - 1);

    verStr = getVersionId();
    verStr += "\n";
    buf    = verStr.getBytes();

    sshOut.write(buf);
    sshOut.flush();
  }

  void receiveServerData() throws IOException {
    SSHPduInputStream pdu = new SSHPduInputStream(SMSG_PUBLIC_KEY, null);
    pdu.readFrom(sshIn);
    int bits;
    BigInteger e, n;

    srvCookie = new byte[8];
    pdu.read(srvCookie, 0, 8);

    bits = pdu.readInt();
    e = pdu.readBigInteger();
    n = pdu.readBigInteger();
    srvServerKey = new KeyPair(new RSAPublicKey(e, n), null);

    bits = pdu.readInt();
    e = pdu.readBigInteger();
    n = pdu.readBigInteger();
    srvHostKey = new KeyPair(new RSAPublicKey(e, n), null);

    if(!authenticator.verifyKnownHosts((RSAPublicKey)srvHostKey.getPublic()))
      throw new IOException("Verification of known hosts failed");

    protocolFlags      = pdu.readInt();
    supportedCiphers   = pdu.readInt();
    supportedAuthTypes = pdu.readInt();
  }

  void generateSessionKey() {
    SecureRandom rand = secureRandom();
    sessionKey = new byte[SESSION_KEY_LENGTH / 8];
    rand.nextBytes(sessionKey);
  }

  void sendSessionKey(int cipherType) throws IOException {
    byte[]             key = new byte[sessionKey.length + 1];
    BigInteger         encKey;
    RSACipher          rsa;
    SSHPduOutputStream pdu;

    key[0] = 0;
    System.arraycopy(sessionKey, 0, key, 1, sessionKey.length);

    for(int i = 0; i < sessionId.length; i++)
      key[i + 1] ^= sessionId[i];

    encKey = new BigInteger(key);

    if(((RSAPublicKey)(srvServerKey.getPublic())).bitLength() <
       ((RSAPublicKey)(srvHostKey.getPublic())).bitLength()) {
      BigInteger padded;
      rsa = new RSACipher(srvServerKey);
      padded = rsa.doPad(encKey, ((RSAPublicKey)srvServerKey.getPublic()).bitLength(), secureRandom());
      encKey = rsa.doPublic(padded);
      rsa = new RSACipher(srvHostKey);
      padded = rsa.doPad(encKey, ((RSAPublicKey)srvHostKey.getPublic()).bitLength(), secureRandom());
      encKey = rsa.doPublic(padded);
    } else {
      BigInteger padded;
      rsa = new RSACipher(srvHostKey);
      padded = rsa.doPad(encKey, ((RSAPublicKey)srvHostKey.getPublic()).bitLength(), secureRandom());
      encKey = rsa.doPublic(padded);
      rsa = new RSACipher(srvServerKey);
      padded = rsa.doPad(encKey, ((RSAPublicKey)srvServerKey.getPublic()).bitLength(), secureRandom());
      encKey = rsa.doPublic(padded);
    }

    pdu = new SSHPduOutputStream(CMSG_SESSION_KEY, null);
    pdu.writeByte((byte)cipherType);
    pdu.write(srvCookie, 0, srvCookie.length);
    pdu.writeBigInteger(encKey);
    // !!! TODO: check this pdu.writeInt(PROTOFLAG_SCREEN_NUMBER | PROTOFLAG_HOST_IN_FWD_OPEN);
    pdu.writeInt(protocolFlags);
    pdu.writeTo(sshOut);

    // !!!
    // At this stage the communication is encrypted
    // !!!

    if(!isSuccess())
      throw new IOException("Error while sending session key!");
  }

  void authenticateUser(String userName) throws IOException {
    SSHPduOutputStream outpdu;

    outpdu = new SSHPduOutputStream(CMSG_USER, sndCipher);
    outpdu.writeString(userName);
    outpdu.writeTo(sshOut);

    if(isSuccess()) {
      user.report("Authenticated directly by server, no other authentication required");
      return;
    }

    int authType = authenticator.getAuthType(user);
    if(!isAuthTypeSupported(authType))
      throw new IOException("Server does not support '" + authTypeDesc[authType] +
			     "' authentication");

    switch(authType) {
    case AUTH_RSA:
      doRSAAuth(false, userName);
      break;
    case AUTH_PASSWORD:
      doPasswdAuth(userName);
      break;
    case AUTH_RHOSTS_RSA:
      doRSAAuth(true, userName);
      break;
    case AUTH_RHOSTS:
    case AUTH_TIS:
    case AUTH_KERBEROS:
    case PASS_KERBEROS_TGT:
      throw new IOException("We do not support selected authentication type " +
			    authTypeDesc[authType]);
    }
  }

  void doPasswdAuth(String userName) throws IOException {
    SSHPduOutputStream outpdu;
    String password;

    password = authenticator.getPassword(user);

    outpdu = new SSHPduOutputStream(CMSG_AUTH_PASSWORD, sndCipher);
    outpdu.writeString(password);
    outpdu.writeTo(sshOut);

    if(!isSuccess())
      throw new IOException("Permission denied");
  }

  void doRSAAuth(boolean rhosts, String userName) throws IOException {
    SSHPduOutputStream outpdu;
    SSHRSAKeyFile      keyFile = authenticator.getIdentityFile(user);
    RSAPublicKey       pubKey  = keyFile.getPublic();

    if(rhosts) {
      outpdu = new SSHPduOutputStream(CMSG_AUTH_RHOSTS_RSA, sndCipher);
      outpdu.writeString(userName);
      outpdu.writeInt(pubKey.bitLength());
      outpdu.writeBigInteger(pubKey.getE());
      outpdu.writeBigInteger(pubKey.getN());
    } else {
      outpdu = new SSHPduOutputStream(CMSG_AUTH_RSA, sndCipher);
      outpdu.writeBigInteger(pubKey.getN());
    }
    outpdu.writeTo(sshOut);

    SSHPduInputStream inpdu = new SSHPduInputStream(MSG_ANY, rcvCipher);
    inpdu.readFrom(sshIn);
    if(inpdu.type == SMSG_FAILURE)
      throw new IOException("Server refused our key" + (rhosts ? " or rhosts" : ""));
    else if(inpdu.type != SMSG_AUTH_RSA_CHALLENGE)
      throw new IOException("Protocol error, expected RSA-challenge but got " + inpdu.type);

    BigInteger challenge = inpdu.readBigInteger();

    // First try with an empty passphrase...
    //
    RSAPrivateKey privKey = keyFile.getPrivate("");
    if(privKey == null)
      privKey = keyFile.getPrivate(authenticator.getIdentityPassword(user));

    if(privKey == null)
      throw new IOException("Invalid password for key-file '" + keyFile.getComment() + "'");

    rsaChallengeResponse(privKey, challenge);
  }

  void rsaChallengeResponse(RSAPrivateKey privKey, BigInteger challenge) throws IOException {
    RSACipher     rsa = new RSACipher(new KeyPair(null, privKey));
    MessageDigest md5;

    challenge = rsa.doPrivate(challenge);
    challenge = rsa.stripPad(challenge);
    byte[] response = challenge.toByteArray();

    try {
      md5 = MessageDigest.getInstance("MD5");
      if(response[0] == 0)
	md5.update(response, 1, 32);
      else
	md5.update(response, 0, 32);
      md5.update(sessionId);
      response = md5.digest();
    } catch(Exception e) {
      throw new IOException("MD5 not implemented, can't generate session-id");
    }

    SSHPduOutputStream outpdu = new SSHPduOutputStream(CMSG_AUTH_RSA_RESPONSE, sndCipher);
    outpdu.write(response, 0, response.length);
    outpdu.writeTo(sshOut);

    if(!isSuccess())
      throw new IOException("Permission denied");
  }

  void initiateSession() throws IOException {
    // !!! TODO: Check if the gzip of java can be used to do this...
    // requestCompression(3);

    if(user.wantPTY())
      requestPTY();

    int maxPktSz = user.getMaxPacketSz();
    if(maxPktSz > 0)
      requestMaxPacketSz(maxPktSz);

    if(user.wantX11Forward())
      requestX11Forward();

    initiateTunnels();

    if(commandLine != null)
      requestCommand(commandLine);
    else
      requestShell();

    // !!!
    // At this stage we can't send more options/forward-requests
    // the server has now entered it's service-loop.
  }

  void initiatePlugins(boolean loadExtra) {
    // !!! TODO: Handle plugins generically through per user properties
    // should only be done once at startup (or when a new plugin is added).
    //
    if(loadExtra) {
      SSHFtpPlugin ftpPlugin = new SSHFtpPlugin();
      SSHProtocolPlugin.addPlugin("ftp", ftpPlugin);
      ftpPlugin.initiate(this);
    }
  }

  void initiateTunnels() throws IOException {
    int i;
    for(i = 0; i < localForwards.size(); i++) {
      LocalForward fwd = (LocalForward) localForwards.elementAt(i);
      requestLocalPortForward(fwd.localPort, fwd.remoteHost, fwd.remotePort, fwd.plugin);
    }
    for(i = 0; i < remoteForwards.size(); i++) {
      RemoteForward fwd = (RemoteForward) remoteForwards.elementAt(i);
      requestRemotePortForward(fwd.remotePort, fwd.localHost, fwd.localPort, fwd.plugin);
    }
  }

  void requestCompression(int level) throws IOException {
    SSHPduOutputStream outpdu = new SSHPduOutputStream(CMSG_REQUEST_COMPRESSION, sndCipher);
    outpdu.writeInt(level);
    outpdu.writeTo(sshOut);
    if(!isSuccess())
      user.report("Error requesting compression level: " + level);
  }

  void requestMaxPacketSz(int sz) throws IOException {
    SSHPduOutputStream outpdu = new SSHPduOutputStream(CMSG_MAX_PACKET_SIZE, sndCipher);
    outpdu.writeInt(sz);
    outpdu.writeTo(sshOut);
    if(!isSuccess())
      user.report("Error requesting max packet size: " + sz);
  }

  void requestX11Forward() throws IOException {
    SSHPduOutputStream outpdu = new SSHPduOutputStream(CMSG_X11_REQUEST_FORWARDING, sndCipher);

    // !!!
    outpdu.writeString("MIT-MAGIC-COOKIE-1");
    outpdu.writeString("112233445566778899aabbccddeeff00");
    outpdu.writeInt(0);
    // !!!

    outpdu.writeTo(sshOut);

    if(!isSuccess())
      user.report("Error requesting X11 forward");
  }

  void requestPTY() throws IOException {
    SSHPduOutputStream outpdu = new SSHPduOutputStream(CMSG_REQUEST_PTY, sndCipher);
    Terminal myTerminal = null;
    if(console != null)
      myTerminal = console.getTerminal();
    if(myTerminal != null) {
      outpdu.writeString(myTerminal.terminalType());
      outpdu.writeInt(myTerminal.rows());
      outpdu.writeInt(myTerminal.cols());
      outpdu.writeInt(myTerminal.vpixels());
      outpdu.writeInt(myTerminal.hpixels());
    } else {
      outpdu.writeString("");
      outpdu.writeInt(0);
      outpdu.writeInt(0);
      outpdu.writeInt(0);
      outpdu.writeInt(0);
    }
    outpdu.writeByte((byte)TTY_OP_END);
    outpdu.writeTo(sshOut);

    if(!isSuccess())
      user.report("Error requesting PTY");
  }

  void requestLocalPortForward(int localPort, String remoteHost, int remotePort, String plugin) {
    controller.newListenChannel(localPort, remoteHost, remotePort, plugin);
  }

  void requestRemotePortForward(int remotePort, String localHost, int localPort, String plugin)
    throws IOException {

    SSHProtocolPlugin.getPlugin(plugin).remoteListener(remotePort, localHost, localPort,
						       controller);

    SSHPduOutputStream outpdu = new SSHPduOutputStream(CMSG_PORT_FORWARD_REQUEST, sndCipher);
    outpdu.writeInt(remotePort);
    outpdu.writeString(localHost);
    outpdu.writeInt(localPort);
    outpdu.writeTo(sshOut);

    if(!isSuccess())
      user.report("Error requesting remote port forward: " + plugin + "/" +remotePort + ":" +
		  localHost + ":" + localPort);
  }

  void requestCommand(String command) throws IOException {
    SSHPduOutputStream outpdu = new SSHPduOutputStream(CMSG_EXEC_CMD, sndCipher);
    outpdu.writeString(command);
    outpdu.writeTo(sshOut);
  }

  void requestShell() throws IOException {
    SSHPduOutputStream outpdu = new SSHPduOutputStream(CMSG_EXEC_SHELL, sndCipher);
    outpdu.writeTo(sshOut);
  }

  boolean isSuccess() throws IOException {
    boolean success = false;
    SSHPduInputStream inpdu = null;
    inpdu = new SSHPduInputStream(MSG_ANY, rcvCipher);
    inpdu.readFrom(sshIn);
    if(inpdu.type == SMSG_SUCCESS)
      success = true;
    else if(inpdu.type == SMSG_FAILURE)
      success = false;
    else if(inpdu.type == MSG_DISCONNECT)
      throw new IOException("Server disconnected: " + inpdu.readString());
    else 
      throw new IOException("Protocol error: got " + inpdu.type + 
			    " when expecting success/failure");
    return success;
  }

  void setInteractive() {
    try {
      sshSocket.setTcpNoDelay(true);
    } catch (SocketException e) {
      user.report("Error setting interactive mode: " + e.getMessage());
    }
  }

  void stdinWriteChar(char c) {
    stdinWriteString(String.valueOf(c));
  }

  void stdinWriteString(String str) {
    try {
      SSHPduOutputStream stdinPdu;
      stdinPdu = new SSHPduOutputStream(SSH.CMSG_STDIN_DATA, sndCipher);
      stdinPdu.writeString(str);
      controller.transmit(stdinPdu);
    } catch (Exception ex) {
      SSH.log("stdinWriteString: " + ex.toString());
    }
  }

  void stdinWriteString(byte[] str) {
    try {
      SSHPduOutputStream stdinPdu;
      stdinPdu = new SSHPduOutputStream(SSH.CMSG_STDIN_DATA, sndCipher);
      stdinPdu.writeInt(str.length);
      stdinPdu.write(str, 0, str.length);
      controller.transmit(stdinPdu);
    } catch (Exception ex) {
      SSH.log("stdinWriteString: " + ex.toString());
    }
  }

  void signalWindowChanged(int rows, int cols, int vpixels, int hpixels) {
    try {
      SSHPduOutputStream pdu;
      pdu = new SSHPduOutputStream(SSH.CMSG_WINDOW_SIZE, sndCipher);
      pdu.writeInt(rows);
      pdu.writeInt(cols);
      pdu.writeInt(vpixels);
      pdu.writeInt(hpixels);
      controller.transmit(pdu);
    } catch (Exception ex) {
      SSH.log("sigWinch: " + ex.toString());
    }
  }

}
