/******************************************************************************
 *
 * Copyright (c) 1998,99 by Mindbright Technology AB, Stockholm, Sweden.
 *                 www.mindbright.se, info@mindbright.se
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 *****************************************************************************
 * $Author: mats $
 * $Date: 1999/02/01 07:27:00 $
 * $Name: rel0-98-1 $
 *****************************************************************************/
package mindbright.ssh;

import java.io.*;
import java.net.*;
import java.util.Properties;
import java.util.Enumeration;
import java.util.NoSuchElementException; 

import mindbright.terminal.*;
import mindbright.security.*;

public final class SSHInteractiveClient extends SSHClient
  implements Runnable, SSHClientUser, SSHAuthenticator {

  static public class DumbConsoleThread implements Runnable {
    SSHChannelController controller;
    SSHStdIO             console;

    public DumbConsoleThread(SSHChannelController controller, SSHStdIO console) {
      this.controller = controller;
      this.console    = console;
    }

    public void run() {
      SSHPduOutputStream stdinPdu;
      String line;
      try {
	while(true) {
	  line = console.promptLine("", "", false);
	  stdinPdu = new SSHPduOutputStream(SSH.CMSG_STDIN_DATA, console.sndCipher);
	  stdinPdu.writeString(line + "\n");
	  controller.transmit(stdinPdu);
	  Thread.sleep(400);
	}
      } catch (SSHStdIO.CtrlDPressedException e) {
	controller.sendDisconnect("exit");
      } catch (Exception e) {
	// !!!
	SSH.log("console-thread: " + e.toString());
      }
    }
  }

  public static String copyright() {
    return "Copyright (c) 1998,99 by Mindbright Technology AB, Stockholm, Sweden";
  }

  Thread        dumbConsoleThread;
  String        sshHomeDir;
  String        knownHosts;
  SSHRSAKeyFile keyFile;
  int           numOfLocalFwd;
  int           numOfRemoteFwd;

  static public final int PROP_NAME    = 0;
  static public final int PROP_VALUE   = 1;
  static public final int PROP_DESC    = 2;
  static public final int PROP_ALLOWED = 3;
  static public final Properties defaultProperties = new Properties();
  static public final String[][] defaultPropDesc = {
    { "server",   "",            "name of server to connect to", "" },
    { "port",     String.valueOf(SSH.DEFAULTPORT),
                                "port on server to connect to", "" },
    { "usrname",  "",           "username to login as", "" },
    { "cipher",   SSH.getCipherName(SSH.CIPHER_DEFAULT),
                                "name of block cipher to use",
                                ("( " + SSH.listSupportedCiphers() + ")") },
    { "authtyp",  "passwd",      "method of authentication",
                                ("( " + SSH.listSupportedAuthTypes() + ")") },
    { "idfile",   "identity",    "name of file containing identity (rsa-keys)", "" },
    { "display",  "localhost:0", "display definition (i.e. <host>:<screen>)", "" },
    { "mtu",      "0",           "maximum packet size to use (0 means use default)",
                                "(4096 - 256k)" },
    { "escseq",   "~$",          "sequence of characters to type to enter local command-shell", "" },
    { "secrand",  "0",           "level of security in random-seed (for generating session-key)",
      "(0-2, 0=low and 2=high)" },
    { "x11fwd",   "false",       "indicates whether X11 display is forwarded or not", "(true/false)" },
    { "prvport",  "false",       "indicates whether to use a privileged port or not (locally)", "(true/false)" },
    { "forcpty",  "true",        "indicates whether to allocate a pty or not", "(true/false)" },
    { "remfwd",   "false",       "indicates whether we allow remote connects to local forwards", "(true/false)" },
    { "idhost",   "true",        "indicates whether to check hosts host-key in 'known_hosts'", "(true/false)" },
    { "portftp",  "false",       "indicates whether to enable ftp 'PORT' command support", "(true/false)" },
  };
  static {
    for(int i = 0; i < defaultPropDesc.length; i++)
      defaultProperties.put(defaultPropDesc[i][PROP_NAME], defaultPropDesc[i][PROP_VALUE]);
  }

  Properties props;
  boolean    propsChanged;
  boolean    setPropPending;
  boolean    isLoadingProperties;
  protected boolean    usePropFiles;
  protected boolean    useDefPropOnly;

  public SSHInteractiveClient(String sshHomeDir, String username) {
    super(null, null); // !!!
    authenticator = this;
    user          = this;

    this.sshHomeDir = sshHomeDir;
    this.knownHosts = KNOWN_HOSTS_FILE;

    setDefaultProperties();
    propsChanged        = false;
    usePropFiles        = true;
    useDefPropOnly      = false;
    setPropPending      = false;
    isLoadingProperties = false;

    numOfLocalFwd  = 0;
    numOfRemoteFwd = 0;

    if(username != null)
      props.put("usrname", username);

    setConsole(new SSHStdIO());
    ((SSHStdIO)console).setClient(this);
  }

  public void printCopyright() {
    console.println(copyright());
  }

  public void run() {
    boolean doCommandShell;
    console.print("Initializing random generator, please wait...");
    try {
	Thread.sleep(100); // !!! let the text show... (?)
    } catch (InterruptedException e) {
	// !!!
    }
    SSH.initSeedGenerator();
    console.println("done");

    boolean keepRunning = true;
    while(keepRunning) {
      doCommandShell = false;
      try {
	console.println("");
	console.println("(press <ctrl> + <mouse3> for main-menu, or");
	console.println(" press <ctrl> + 'D' to enter local command-shell)");
	if(isDumb())
	  console.println("(...you might have to press ENTER also...");
	console.println("");

	// This starts a connection to the sshd and all the related stuff...
	//
	bootSSH(true);

	if(isDumb())
	  startDumbConsole();

	// Join main-receiver channel thread and wait for session to end
	//
	controller.waitForExit();

	if(isDumb())
	  stopDumbConsole();

	// !!! Wait for last session to close down entirely (i.e. so
	// disconnected gets a chance to be called...)
	Thread.sleep(1000);

      } catch(SSHStdIO.CtrlDPressedException e) {
	doCommandShell = true;
      } catch(UnknownHostException e) {
	console.println("Unknown host: " + getProperty("server"));
      } catch(FileNotFoundException e) {
	console.println("File not found: " + e.getMessage());
      } catch(Exception e) {
	console.println("Error connecting to " + getProperty("server") + ", reason:");
	console.println("-> " + e.getMessage());
      } catch(ThreadDeath death) {
	if(controller != null)
	  controller.killAll();
	if(SSH.bogusThread != null && SSH.bogusThread.isAlive())
	  bogusThread.stop();
	throw death;
      }

      console.println("");

      if(doCommandShell) {
	keepRunning = ((SSHStdIO)console).doCommandShell();
      }
    }
  }

  public boolean isDumb() {
    return (console.getTerminal() == null);
  }

  public TerminalWin getTerminalWin() {
    Terminal term = console.getTerminal();
    if(term != null && term instanceof TerminalWin)
      return (TerminalWin)term;
    return null;
  }

  public void startDumbConsole() {
    Runnable dumbConsole = new DumbConsoleThread(controller, (SSHStdIO)console);
    dumbConsoleThread = new Thread(dumbConsole);
    dumbConsoleThread.start();
  }
  public void stopDumbConsole() {
    dumbConsoleThread.stop();
  }

  public String promptLine(String prompt, String defaultVal, boolean echoStar) throws IOException {
    return ((SSHStdIO)console).promptLine(prompt, defaultVal, echoStar);
  }
  public void updateTitle() {
    ((SSHStdIO)console).updateTitle();
  }

  public void setSSHHomeDir(String sshHomeDir) {
    if(!sshHomeDir.endsWith(File.separator))
      sshHomeDir += File.separator;
    this.sshHomeDir = sshHomeDir;
  }
  public String getSSHHomeDir() {
    return sshHomeDir;
  }

  //
  // Methods delegated to Properties
  //
  public void setUsePropertyFiles(boolean val) {
    usePropFiles = val;
  }
  public void setUseOnlyDefaultProperties(boolean val) {
    useDefPropOnly = val;
  }

  public Properties mergeProperties(Properties props1, Properties props2) {
    Properties  merged = new Properties();
    Enumeration enum;
    String      name;

    enum = props1.propertyNames();
    while(enum.hasMoreElements()) {
      name = (String)enum.nextElement();
      merged.put(name, props1.getProperty(name));
    }

    if(props2 != null) {
      enum = props2.propertyNames();
	while(enum.hasMoreElements()) {
	  name = (String)enum.nextElement();
	  merged.put(name, props2.getProperty(name));
	}
    }

    return merged;
  }

  public void setDefaultProperties() {
    props = mergeProperties(defaultProperties, null);
  }

  public boolean loadDefaultProperties() {
    return loadProperties(PROPS_FILE_DEF);
  }

  public boolean saveDefaultProperties() {
    return saveProperties(PROPS_FILE_DEF);
  }

  public boolean saveCurrentProperties() {
    boolean ret = true;

    if(useDefPropOnly) {
      if(propsChanged) {
	ret = saveProperties(PROPS_FILE_DEF);
	if(ret)
	  console.println("Saved default property-file.");
      }
    } else {
      String srv = getProperty("server");
      if(srv != null && !srv.equals("") && propsChanged) {
	ret = saveProperties(srv);
	if(ret)
	  console.println("Saved property-file '" + sshHomeDir + srv + PROPS_FILE_EXT + "'.");
      }
    }
    return ret;
  }

  public boolean loadProperties(String fname) {
    if(!usePropFiles)
      return true;
    FileInputStream f;
    TerminalWin term = getTerminalWin();
    try {
      f = new FileInputStream(sshHomeDir + fname + PROPS_FILE_EXT);
      Properties loadProps = new Properties();
      loadProps.load(f);

      Enumeration enum;
      String      name;

      isLoadingProperties = true;
      numOfLocalFwd       = 0;
      numOfRemoteFwd      = 0;

      enum = loadProps.propertyNames();
      while(enum.hasMoreElements()) {
	name = (String)enum.nextElement();
	try {
	  try {
	    setProperty(name, loadProps.getProperty(name));
	  } catch (NoSuchElementException e1) {
	    if(term != null) {
	      try {
		term.setProperty(name, loadProps.getProperty(name));
	      } catch (NoSuchElementException e2) {
		console.println("No such property: " + name);
	      }
	    }
	  }
	} catch (IllegalArgumentException e3) {
	  console.println(e3.getMessage());
	}
      }

    } catch (IOException e) {
      return false;
    } finally {
      isLoadingProperties = false;
    }
    return true;
  }

  public boolean saveProperties(String fname) {
    if(!usePropFiles)
      return true;
    FileOutputStream f;
    Properties       termProps = (getTerminalWin() != null ? getTerminalWin().getProperties() :
				  null);
    try {
      f = new FileOutputStream(sshHomeDir + fname + PROPS_FILE_EXT);
      props.save(f, "MindTerm ssh-properties");
      if(termProps != null)
	termProps.save(f, "MindTerm terminal-properties");
      f.close();
    } catch (IOException e) {
      return false;
    }
    return true;
  }

  public String getProperty(String key) {
    return props.getProperty(key);
  }

  public void setProperty(String key, String value) throws IllegalArgumentException,
    NoSuchElementException {
    boolean setChanged  = !(value.equals(getProperty(key)));
    boolean forceChange = false;
    //
    // Some sanity checks...
    //
    if(key.equals("cipher")) {
      if(SSH.getCipherType(value) == SSH.CIPHER_NOTSUPPORTED)
	throw new IllegalArgumentException("cipher " + value + " not supported");
      //
    } else if(key.equals("authtyp")) {
      if(SSH.getAuthType(value) == AUTH_NOTSUPPORTED)
	throw new IllegalArgumentException("authtype " + value + " not supported");
      //
    } else if(key.equals("x11fwd")  || key.equals("prvport") ||
	      key.equals("forcpty") || key.equals("remfwd")   ||
	      key.equals("idhost")  || key.equals("portftp")) {
      if(!(value.equals("true") || value.equals("false")))
	throw new IllegalArgumentException("value for " + key + " must be 'true' or 'false'");
      if(key.equals("remfwd"))
	SSHListenChannel.setAllowRemoteConnect((new Boolean(value)).booleanValue());
      else if(key.equals("portftp"))
	SSHFtpPlugin.supportPORT = (new Boolean(value)).booleanValue();
      //
    } else if(key.equals("port") || key.equals("mtu") || key.equals("secrand")) {
      try {
	int val = Integer.valueOf(value).intValue();
	if(key.equals("port") && (val > 65535 || val < 0)) {
	  throw new IllegalArgumentException("not a valid port number: " + value);
	} else if(key.equals("mtu") && val != 0 && (val > (256*1024) || val < 4096)) {
	  throw new IllegalArgumentException("mtu must be between 4k and 256k");
	} else if(key.equals("secrand")) {
	  if(val < 0 || val > 2)
	    throw new IllegalArgumentException("secrand must be 0-2");
	  else
	    SecureRandom.secureLevel = val;
	}
      } catch (NumberFormatException e) {
	throw new IllegalArgumentException("value for " + key + " must be an integer");
      }
      //
    } else if(key.equals("server")) {
      if(!setPropPending) {
	if(isConnected) {
	  throw new IllegalArgumentException("server can only be set while not connected");
	} else {
	  String oldHost = getProperty("server");
	  if(!oldHost.equals(value)) {
	    if(!useDefPropOnly) {
	      if(!saveCurrentProperties())
		console.println("Error saving property-file for '" + oldHost + "'.");
	      setPropPending = true; // !!! We will call setProperty() in loadProperties()
	      if(!loadProperties(value)) {
		console.println("No property-file found for '" + value + "', will be created later.");
	      } else {
		forceChange = true;
		setChanged  = false;
	      }
	      setPropPending = false;
	    }
	  } else {
	    setChanged  = false;
	  }
	}
      }
    } else if(key.equals("usrname") || key.equals("idfile")  ||
	      key.equals("display") || key.equals("escseq")) {
      // !!! Do nothing for these, arbitrary strings...
      //
    } else if(key.startsWith("local")) {
      try {
	int n = Integer.parseInt(key.substring(5));
	if(!isLoadingProperties && n > numOfLocalFwd)
	  throw new IllegalArgumentException("port forwards must be given in sequence");
	if(isLoadingProperties) {
	  numOfLocalFwd++;
	} else if(n == numOfLocalFwd) {
	  numOfLocalFwd++;
	}
	addLocalPortForward(value, true);
      } catch (Exception e) {
	throw new IllegalArgumentException("not a valid port forward: " + key + " : " + value);
      }
    } else if(key.startsWith("remote")) {
      try {
	int n = Integer.parseInt(key.substring(6));
	if(!isLoadingProperties && n > numOfRemoteFwd)
	  throw new IllegalArgumentException("port forwards must be given in sequence");
	if(isLoadingProperties) {
	  numOfRemoteFwd++;
	} else if(n == numOfRemoteFwd) {
	  numOfRemoteFwd++;
	}
	addRemotePortForward(value, true);
      } catch (Exception e) {
	throw new IllegalArgumentException("not a valid port forward: " + key + " : " + value);
      }
    } else {
      throw new NoSuchElementException("unknown ssh-property '" + key + "'");
    }

    if(!propsChanged || forceChange)
      propsChanged = setChanged;

    props.put(key, value);
    updateTitle();
  }

  public Enumeration propertyNames() {
    return props.propertyNames();
  }

  public int getNumOfLocalForwards() {
    return numOfLocalFwd;
  }

  public int getNumOfRemoteForwards() {
    return numOfRemoteFwd;
  }

  // SSHAuthenticator interface
  //
  public String getUsername(SSHClientUser origin) throws IOException {
    String username = getProperty("usrname");
    if(commandLine == null || username == null) {
      username = promptLine(getProperty("server") + " login: ", username, false);
      setProperty("usrname", username); // Changing the user-name does not save new properties...
    }
    return username;
  }

  public String getPassword(SSHClientUser origin) throws IOException {
    String password;
    password = promptLine(getProperty("usrname") + "@" + getProperty("server") +
			  "'s password: ", "", true);
    return password;
  }

  public int getAuthType(SSHClientUser origin) {
    return SSH.getAuthType(getProperty("authtyp"));
  }

  public int getCipher(SSHClientUser origin) {
    return SSH.getCipherType(getProperty("cipher"));
  }

  public SSHRSAKeyFile getIdentityFile(SSHClientUser origin) throws IOException {
    keyFile = new SSHRSAKeyFile(sshHomeDir + File.separator +
				getProperty("idfile"));
    return keyFile;
  }
  
  public String getIdentityPassword(SSHClientUser origin) throws IOException {
    String password;
    password = promptLine("key-file '" + keyFile.getComment() + "' password: ", "", true);
    return password;
  }

  public boolean verifyKnownHosts(RSAPublicKey hostPub) throws IOException {
    if(!Boolean.valueOf(getProperty("idhost")).booleanValue())
      return true;

    int    hostCheck;
    String answer;
    String fileName = sshHomeDir + File.separator + knownHosts;
    File   tmpFile = new File(fileName);

    if(!tmpFile.exists()) {
      answer = promptLine("File '"  + fileName + "' not found, do you want to create it (yes/no)? ", "", false);
      if(answer.equalsIgnoreCase("yes") || answer.equalsIgnoreCase("y")) {
	FileOutputStream f = new FileOutputStream(tmpFile);
	f.close();
      } else {
	throw new IOException("When 'idhost' is true, file 'known_hosts' must be accesible for read/write");
      }
    }

    SSHRSAPublicKeyFile file = new SSHRSAPublicKeyFile(fileName, true);

    if((hostCheck = file.checkPublic(hostPub.getN(), getProperty("server"))) ==
       SRV_HOSTKEY_KNOWN)
      return true;

    if(hostCheck == SRV_HOSTKEY_NEW) {
	console.println("Host key not found from the list of known hosts.");
	answer = promptLine("Do you want to add this host to your set of known hosts (yes/no)? ",
			    "", false);
    } else {
	console.println("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@");
	console.println("@       WARNING: HOST IDENTIFICATION HAS CHANGED!         @");
	console.println("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@");
	console.println("IT IS POSSIBLE THAT SOMEONE IS DOING SOMETHING NASTY,");
	console.println("ONLY PROCEED IF YOU KNOW WHAT YOU ARE DOING!");
	answer = promptLine("Do you want to replace the identification of this host (yes/no)? ",
			    "", false);
	file.removePublic(getProperty("server"));
    }

    if(answer.equalsIgnoreCase("yes") || answer.equalsIgnoreCase("y")) {
      file.addPublic(getProperty("server"), null, hostPub.getE(), hostPub.getN());
      tmpFile      = new File(fileName + ".tmp");
      File oldFile = new File(fileName);
      oldFile.renameTo(tmpFile);
      try {
	file.saveToFile(fileName);
      } catch (IOException e) {
	oldFile = new File(fileName);
	tmpFile.renameTo(oldFile);
	throw e;
      }
      tmpFile.delete();
    } else {
      return false;
    }

    return true;
  }

  //
  // SSHClientUser interface
  //

  public String getSrvHost() throws IOException {
    String host = getProperty("server");
    if(commandLine == null || host == null) {
      host = promptLine("SSH-server: ", host, false);
      setProperty("server", host);
    }
    return host;
  }

  public int getSrvPort() {
    return Integer.valueOf(getProperty("port")).intValue();
  }

  public String getDisplay() {
    return getProperty("display");
  }

  public int getMaxPacketSz() {
    return Integer.valueOf(getProperty("mtu")).intValue();
  }

  public boolean wantX11Forward() {
    return Boolean.valueOf(getProperty("x11fwd")).booleanValue();
  }

  public boolean wantPrivileged() {
    return Boolean.valueOf(getProperty("prvport")).booleanValue();
  }

  public boolean wantPTY() {
    return Boolean.valueOf(getProperty("forcpty")).booleanValue();
  }

  public void connected(SSHClient client) {
    SSHMenuHandler.updateMenus();
    updateTitle();
    console.println("Connected to server running " + srvVersion);
    console.println("(command-shell escape-sequence is '" + ((SSHStdIO)console).escapeString + "')");
    console.println("");
  }

  public void disconnected(SSHClient client, boolean graceful) {
    SSHMenuHandler.updateMenus();
    srvVersion = null;
    updateTitle();
  }

  public void report(String msg) {
    console.println(msg);
    console.println("");
  }
}
