/******************************************************************************
 *
 * Copyright (c) 1998,99 by Mindbright Technology AB, Stockholm, Sweden.
 *                 www.mindbright.se, info@mindbright.se
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 *****************************************************************************
 * $Author: mats $
 * $Date: 1999/01/10 15:15:37 $
 * $Name: rel0-98-1 $
 *****************************************************************************/
package mindbright.ssh;

import java.io.*;
import java.math.BigInteger;

import mindbright.security.*;
import mindbright.util.*;

public class SSHPduOutputStream extends SSHDataOutputStream implements SSHPdu {

  static class PduByteArrayOutputStream extends ByteArrayOutputStream {
    PduByteArrayOutputStream() {
      super();
    }

    PduByteArrayOutputStream(int size) {
      super(size);
    }

    public byte[] getBuf() {
      return buf;
    }

    public int getCount() {
      return count;
    }

    public void setBuf(byte[] buf) {
      this.buf = buf;
    }

    public void setCount(int count) {
      this.count = count;
    }
  }

  public static final int SSH_DEFAULT_PKT_LEN = 8192;
  public static int mtu = SSH_DEFAULT_PKT_LEN;

  public static synchronized void setMTU(int newMtu) {
    mtu = newMtu;
  }

  public static synchronized int getMTU() {
    return mtu;
  }

  byte[]  readFromRawData;
  int     readFromOff;
  int     readFromSize;

  public int    type;
  public Cipher cipher;

  SSHPduOutputStream(Cipher cipher) {
    super(null);
    this.cipher = cipher;
  }

  SSHPduOutputStream(int type, Cipher cipher) throws IOException {
    super(new PduByteArrayOutputStream(getMTU()));
    this.type   = type;
    this.cipher = cipher;
    if(cipher != null) {
      SecureRandom rand = SSH.secureRandom();
      byte[] randBytes  = new byte[8];
      rand.nextBytes(randBytes);
      for(int i = 0; i < 8; i++)
	write(randBytes[i]);
    } else {
      for(int i = 0; i < 8; i++)
	write(0);
    }
    write(type);
  }

  public SSHPdu createPdu() {
    SSHPdu pdu;
    try {
      pdu = new SSHPduOutputStream(this.type, this.cipher);
    } catch (IOException e) {
      // !!!
      pdu = null;
    }
    return pdu;
  }

  public void readFrom(InputStream in) throws IOException {
    if(type != SSH.MSG_CHANNEL_DATA &&
       type != SSH.CMSG_STDIN_DATA)
      throw new IOException("Trying to read raw data into non-data PDU");

    PduByteArrayOutputStream bytes = (PduByteArrayOutputStream) out;

    readFromRawData = bytes.getBuf();
    readFromOff     = bytes.size() + 4; // Leave space for size

    readFromSize = in.read(readFromRawData, readFromOff, getMTU() - readFromOff);
    if(readFromSize == -1)
      throw new IOException("EOF");

    writeInt(readFromSize);
    bytes.setCount(readFromOff + readFromSize);
  }

  public void writeTo(OutputStream sshOut) throws IOException {
    PduByteArrayOutputStream bytes   = (PduByteArrayOutputStream) out;
    int                      pad     = (bytes.size() + 4) % 8;
    byte[]                   padData = new byte[bytes.size() + 4 - pad];
    int                      crc32   = (int)CRC32.getValue(bytes.getBuf(), pad, bytes.size() - pad);

    writeInt(crc32);
    System.arraycopy(bytes.getBuf(), pad, padData, 0, padData.length);

    if(cipher != null) {
      padData = cipher.encrypt(padData);
    }

    bytes = new PduByteArrayOutputStream();
    bytes.setBuf(padData);
    bytes.setCount(padData.length);
    out = bytes;
    sshOut = new SSHDataOutputStream(sshOut);

    ((SSHDataOutputStream)sshOut).writeInt(bytes.getCount() - (8 - pad));
    bytes.writeTo(sshOut);
    sshOut.flush();
  }

  public byte[] rawData() {
    return readFromRawData;
  }
  public void rawSetData(byte[] raw) {
  }
  public int rawOffset() {
    return readFromOff;
  }
  public int rawSize() {
    // !!! return readFromSize;
    //
    byte[] bytes = readFromRawData;
    int    off   = readFromOff - 4;
    int ch1 = ((bytes[off++] + 256) & 0xff);
    int ch2 = ((bytes[off++] + 256) & 0xff);
    int ch3 = ((bytes[off++] + 256) & 0xff);
    int ch4 = ((bytes[off]   + 256) & 0xff);
    return ((ch1 << 24) + (ch2 << 16) + (ch3 << 8) + (ch4 << 0));
  }
  public void rawAdjustSize(int size) {
    PduByteArrayOutputStream bytes = (PduByteArrayOutputStream) out;
    bytes.setCount(size);
  }

}

