/******************************************************************************
 *
 * Copyright (c) 1998,99 by Mindbright Technology AB, Stockholm, Sweden.
 *                 www.mindbright.se, info@mindbright.se
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 *****************************************************************************
 * $Author: mats $
 * $Date: 1999/02/01 07:27:00 $
 * $Name: rel0-98-1 $
 *****************************************************************************/
package mindbright.ssh;

import java.io.*;
import java.awt.*;
import java.awt.datatransfer.*;

import mindbright.terminal.*;
import mindbright.security.*;

public final class SSHStdIO implements TerminalListener, SSHConsole {

  static public class CtrlDPressedException extends IOException {
  }

  static Toolkit toolkit = Toolkit.getDefaultToolkit();

  Container            ownerContainer;
  SSHChannelController controller;
  SSHInteractiveClient client;
  Terminal             term;
  Cipher               sndCipher;
  String               ownerName;

  int                  escapeIdx;
  String               escapeString = "~$";
  boolean              escapeForced = false;

  Boolean readLineLock;
  boolean pressedCtrlD;
  boolean isReadingLine;
  boolean echoStar;
  String  readLineStr;

  protected boolean isConnected;

  public SSHStdIO() {
    this.readLineLock = new Boolean(false);
    this.controller   = null;
    this.sndCipher    = null;
    this.isConnected  = false;
  }

  public boolean isConnected() {
    return isConnected;
  }

  public void setTerminal(Terminal term) {
    this.term = term;
    if(term != null)
      term.addTerminalListener(this);
  }

  public void setClient(SSHInteractiveClient client) {
    this.client = client;
  }

  public void setOwnerContainer(Container ownerContainer) {
    this.ownerContainer = ownerContainer;
  }

  public void setOwnerName(String ownerName) {
    this.ownerName = ownerName;
  }

  public String readLine(String defaultVal) {
    synchronized(readLineLock) {
      if(defaultVal != null) {
	readLineStr   = defaultVal;
	term.write(defaultVal);
      } else {
	readLineStr   = "";
      }
      isReadingLine = true;
      try {
	readLineLock.wait();
      } catch (InterruptedException e) {
	// !!!
      }
      isReadingLine = false;
    }
    return readLineStr;
  }

  public String promptLine(String prompt, String defaultVal, boolean echoStar) throws IOException {
    String line = null;
    pressedCtrlD = false;
    if(term != null) {
      term.setAttribute(Terminal.ATTR_BOLD, true);
      term.write(prompt);
      term.setAttribute(Terminal.ATTR_BOLD, false);
      this.echoStar = echoStar;
      line = readLine(defaultVal);
      this.echoStar = false;
    } else {
      BufferedReader br = new BufferedReader(new InputStreamReader(System.in));
      System.out.print(prompt);
      line = br.readLine();

      // !!! kludge, winlost doesn't give empty input with ^D, this is a kludgy thing
      //  anyway since we're not sure that ^D is EOF in the shell...
      //
      if(line == null || line.charAt(0) == (char)4)
	pressedCtrlD = true;
    }
    if(pressedCtrlD) {
      pressedCtrlD = false;
      throw new CtrlDPressedException();
    }

    return line;
  }

  public void updateTitle() {
    int rows = 0;
    int cols = 0;

    if(term == null || ownerContainer == null)
      return;

    String title = null;
    if(term instanceof TerminalWin)
      title = ((TerminalWin)term).getTitle();

    if(title == null) {
      rows = term.rows();
      cols = term.cols();

      title = ownerName + " - ";
      if(client.getProperty("usrname") != null)
	title += client.getProperty("usrname");
      if(client.getProperty("server") != null) {
	title += "@" + client.getProperty("server");
	if(client.getServerVersion() != null)
	  title += " <" + client.getServerVersion() + ">";
	else
	  title += " <not connected>";
      } else
	title += " <server not known>";
      title += " [" + cols + "x" + rows + "]";
    }
    if(ownerContainer instanceof Frame)
      ((Frame)ownerContainer).setTitle(title);
  }

  // One hell of a kludge all this... (a bit better now at least...)
  //
  public String getNextArg(String args) {
    int i = args.indexOf(' ');
    if(i > -1)
      args = args.substring(0, i);
    return args;
  }
  public String[] makeArgv(String cmdLine) {
    String[] argv = new String[32];
    String[] argvRet;
    int n = 0, i;
    while(cmdLine != null) {
      argv[n++] = getNextArg(cmdLine);
      i = cmdLine.indexOf(' ');
      if(i > -1) {
	cmdLine = cmdLine.substring(i);
	cmdLine = cmdLine.trim();
      } else
	cmdLine = null;
    }
    argvRet = new String[n];
    System.arraycopy(argv, 0, argvRet, 0, n);
    return argvRet;
  }
  public void doHelp() {
    println("The following commands are available:");
    println("");
    println("go                                    Start SSH-session with current settings.");
    println("quit                                  Quit program.");
    println("add <l|r> [/<plug>/]<port>:<host>:<port>  (see below).");
    println("del <l|r> <listen-port>|*             Delete local/remote forward (* = all).");
    println("list [ssh | term]                     Lists ssh- and/or terminal-settings.");
    println("set [<parameter> <value>]             Set value of a ssh-parameter.");
    println("tset [<parameter> <value>]            Set value of a terminal-parameter.");
    println("help                                  Display this list, but you knew that :-).");
    println("");
    println("(do 'set' without arguments to list parameter-usage)");
    println("");
    println("Examples of adding a remote/local tunnel:");
    println("> add r 4711:www.foo.com:80           Adds a remote tunnel at port 4711 back to");
    println("                                      www.foo.com port 80 without a plugin,");
    println("                                      i.e. default tunneling behaviour.");
    println("> add l /ftp/4711:ftp.bar.com:21      Adds a local tunnel going to ftp.bar.com");
    println("                                      port 21 using the ftp protocol-plugin to");
    println("                                      handle protocol specific needs.");
    println("");
    println("NOTE: The first character of the command can be used instead of the full word.");
    println("NOTE2: Sometimes ftp-forward to localhost/127.0.0.1 fails, use hostname instead.");
    println("");
  }
  public void doHelpSet() {
    int i;
    println("SSH-parameters:");
    println("");
    for(i = 0; i < client.defaultPropDesc.length; i++) {
      println(client.defaultPropDesc[i][client.PROP_NAME] + "\t: " +
	      client.defaultPropDesc[i][client.PROP_DESC]);
    }
    println("(to see possible parameter-values use 'list')");
  }
  public void doHelpTSet() {
    int i;
    println("Terminal-parameters:");
    println("");
    for(i = 0; i < TerminalWin.defaultPropDesc.length; i++) {
      println(TerminalWin.defaultPropDesc[i][TerminalWin.PROP_NAME] + "\t: " +
	      TerminalWin.defaultPropDesc[i][TerminalWin.PROP_DESC]);
    }
    println("(to see possible parameter-values use 'list')");
  }
  public void doAdd(String[] argv) {
    if(argv.length < 3 || (!argv[1].equals("l") && !argv[1].equals("r")))
      doHelp();
    else {
      try {
	if(argv[1].equals("l"))
	  client.setProperty("local" + client.getNumOfLocalForwards(), argv[2]);
	else
	  client.setProperty("remote" + client.getNumOfRemoteForwards(), argv[2]);
      } catch (Exception e) {
	doHelp();
      }
    }
  }
  public void doDel(String[] argv) {
    if(argv.length < 3 || (!argv[1].equals("l") && !argv[1].equals("r")))
      doHelp();
    else {
      try {
	int port;
	if(argv[2].equals("*"))
	  port = -1;
	else
	  port = Integer.parseInt(argv[2]);
	if(argv[1].equals("l"))
	  client.delLocalPortForward(port);
	else
	  client.delRemotePortForward(port);
      } catch (Exception e) {
	doHelp();
      }
    }
  }
  public void doListSSH() {
    int i;
    println("");
    if(term != null)
      term.setAttribute(Terminal.ATTR_BOLD, true);
    println("SSH settings:");
    if(term != null)
      term.setAttribute(Terminal.ATTR_BOLD, false);

    for(i = 0; i < client.defaultPropDesc.length; i++) {
      String propName = client.defaultPropDesc[i][client.PROP_NAME];
      String propVal  = client.getProperty(propName);
      println(propName + "\t: " + (propVal.equals("") ? "<not set>" : propVal) + " " +
	      client.defaultPropDesc[i][client.PROP_ALLOWED]);
    }

    println("");
    println("local tunnels:");
    for(i = 0; i < client.localForwards.size(); i++) {
      SSHClient.LocalForward fwd = (SSHClient.LocalForward) client.localForwards.elementAt(i);
      println("\tlocal:  " + fwd.localPort + "\tremote: " + fwd.remoteHost + "/" +
	      fwd.remotePort + " (plugin: " + fwd.plugin + ")");
    }
    if(i == 0)
      println("\t<none>");
    println("remote tunnels:");
    for(i = 0; i < client.remoteForwards.size(); i++) {
      SSHClient.RemoteForward fwd = (SSHClient.RemoteForward) client.remoteForwards.elementAt(i);
      println("\tremote: " + fwd.remotePort + "\tlocal:  " + fwd.localHost + "/" +
	      fwd.localPort + " (plugin: " + fwd.plugin + ")");
    }
    if(i == 0)
      println("\t<none>");
    println("");

    if(client.isConnected) {
      if(term != null)
	term.setAttribute(Terminal.ATTR_BOLD, true);
      println("Currently active tunnels:");
      if(term != null)
	term.setAttribute(Terminal.ATTR_BOLD, false);
      String list = controller.listTunnels();
      if(list.equals(""))
	print("<none>");
      else
	print(list);
      println("");
    }
  }

  public void doListTerm() {
    int i;
    println("");

    if(term != null) {
      TerminalWin termwin = (TerminalWin)term;
      term.setAttribute(Terminal.ATTR_BOLD, true);
      println("Terminal settings:");
      term.setAttribute(Terminal.ATTR_BOLD, false);

      for(i = 0; i < TerminalWin.defaultPropDesc.length; i++) {
	String propName = TerminalWin.defaultPropDesc[i][TerminalWin.PROP_NAME];
	println(propName + "\t: " + termwin.getProperty(propName) + " " +
		TerminalWin.defaultPropDesc[i][TerminalWin.PROP_ALLOWED]);
      }

      println("");
    }
  }

  public void doSet(String[] argv) {
    if(argv.length < 3) {
      doHelpSet();
    } else {
      try {
	String prm = argv[1];
	String arg = argv[2];
	client.setProperty(prm, arg);
      } catch (Exception e) {
	println(e.getMessage());
	println("(use 'set' without parameters to get help on available parameters)");
      }
    }
  }
  public void doTSet(String[] argv) {
    if(argv.length < 3) {
      doHelpTSet();
    } else {
      String prm = argv[1];
      String arg = argv[2];
      TerminalWin termwin = null;
      if(term instanceof TerminalWin)
	termwin = (TerminalWin)term;
      if(termwin != null) {
	try {
	  termwin.setProperty(prm, arg);
	} catch (Exception e) {
	  println(e.getMessage());
	  println("(use 'tset' without parameters to get help on available parameters)");
	}
      } else {
	println("Can't set terminal-parameters in dumb-console mode.");
      }
    }
  }

  public boolean doCommandShell() {
    boolean retVal = true;
    println("");
    println("...entering local command-shell (type 'h' for help).");
    println("");
    try {
      boolean keepRunning = true;
      String[] argv;
      while(keepRunning) {
	String cmdLine = promptLine("mindterm> ", null, false);
	String cmd;
	cmdLine = cmdLine.toLowerCase();
	cmdLine = cmdLine.trim();
	if(cmdLine.equals(""))
	  continue;
	argv = makeArgv(cmdLine);
	cmd  = argv[0];
	if(cmd.equals("l") || cmd.equals("list")) {
	  if(argv.length > 1) {
	    if(argv[1].equals("ssh") || argv[1].equals("term")) {
	      if(argv[1].equals("ssh"))
		doListSSH();
	      else
		doListTerm();
	      println("(permitted values are in parentheses on the right)");
	      println("");
	    } else
	      println("usage: 'list [ssh | term]'");
	  } else {
	    doListSSH();
	    doListTerm();
	    println("(permitted values are in parentheses on the right)");
	    println("");
	  }
	} else if(cmd.equals("a") || cmd.equals("add")) {
	doAdd(argv);
	} else if(cmd.equals("d") || cmd.equals("del")) {
	  doDel(argv);
	} else if(cmd.equals("s") || cmd.equals("set")) {
	  doSet(argv);
	} else if(cmd.equals("t") || cmd.equals("ts") || cmd.equals("tset")) {
	  doTSet(argv);
	} else if(cmd.equals("help") || cmd.equals("?") || cmd.equals("h")) {
	  doHelp();
	} else if(cmd.equals("go") || cmd.equals("g")) {
	  retVal      = true;
	  keepRunning = false;
	} else if(cmd.equals("q") || cmd.equals("quit")) {
	  retVal      = false;
	  keepRunning = false;
	} else {
	  doHelp();
	}
      }
    } catch (CtrlDPressedException e) {
      retVal = true;
    } catch (Exception e) {
      retVal = false;
    }
    return retVal;
  }

  void launchCommandShell() {
    (new Thread(new Runnable() {
      public void run() { 
	isConnected = false;
	if(!doCommandShell()) {
	  controller.sendDisconnect("exit");
	} else {
	  isConnected = true;
	  typedChar((char)0x0c);
	}
      }
    })).start();
  }

  public void wantCommandShell() {
    escapeForced = true;
  }

  boolean escapeSequenceTyped(char c) {
    if(escapeForced) {
      escapeForced = false;
      return true;
    }
    if(c == escapeString.charAt(escapeIdx))
      escapeIdx++;
    else
      escapeIdx = 0;
    if(escapeIdx == escapeString.length()) {
      escapeIdx = 0;
      return true;
    }
    return false;
  }

  // SSHConsole interface
  //
  public Terminal getTerminal() {
    return term;
  }
  public void stdoutWriteString(String str) {
    if(isConnected)
      print(str);
  }
  public void stderrWriteString(String str) {
    if(isConnected)
      print(str);
  }
  public void print(String str) {
    if(term != null) {
      term.write(str);
    } else {
      System.out.print(str);
    }
  }
  public void println(String str) {
    if(term != null) {
      term.write(str + "\n\r");
    } else {
      System.out.println(str);
    }
  }
  public void serverConnect(SSHChannelController controller, Cipher sndCipher) {
    this.controller  = controller;
    this.sndCipher   = sndCipher;
    isConnected      = true;
  }
  public void serverDisconnect(String reason) {
    this.controller  = null;
    this.sndCipher   = null;
    isConnected      = false;
    println(reason);
  }

  // TerminalListener interface
  //
  public void typedChar(char c) {
    if(isConnected) {
      if(escapeSequenceTyped(c)) {
	launchCommandShell();
      } else {
	client.stdinWriteChar(c);
      }
    } else {
      synchronized(readLineLock) {
	if(isReadingLine) {
	  if(c == (char)4) {
	    // Pressed ctrl-d
	    pressedCtrlD = true;
	    readLineLock.notify();
	  } else if(c == (char)127 || c == (char)0x08) {
	    if(readLineStr.length() > 0) {
	      readLineStr = readLineStr.substring(0, readLineStr.length() - 1);
	      term.write((char)8);
	      term.write(' ');
	      term.write((char)8);
	    } else
	      term.doBell();
	  } else if(c == '\r') {
	    //	    readLineStr = readLineStr + "\r";
	    readLineLock.notify();
	    term.write("\n\r");
	  } else {
	    readLineStr = readLineStr + c;
	    if(echoStar)
	      term.write('*');
	    else
	      term.write(c);
	  }
	}
      }
    }
  }
  public void sendBytes(byte[] b) {
    if(isConnected) {
      client.stdinWriteString(b);
    } else {
      for(int i = 0; i < b.length; i++)
	typedChar((char)b[i]);
    }
  }
  public void signalWindowChanged(int rows, int cols, int vpixels, int hpixels) {
    if(isConnected) {
      client.signalWindowChanged(rows, cols, vpixels, hpixels);
    } else {
      // !!!
    }
    updateTitle();
  }
  public void setSelection(String selection) {
    Clipboard cb = ownerContainer.getToolkit().getSystemClipboard();
    if(cb == null || term == null)
      return;
    StringSelection sl = new StringSelection(selection);
    cb.setContents(sl, sl);
  }
  public String getSelection() {
    Clipboard cb = toolkit.getSystemClipboard();
    String    sl = null;
    if(cb == null || term == null)
      return sl;
    Transferable t = cb.getContents(this);
    if(t != null) {
      try {
	sl = (String) t.getTransferData(DataFlavor.stringFlavor);
      } catch (Exception e) {
	toolkit.beep();
      }
    } else
      toolkit.beep();
    return sl;
  }

}
